(*
  6:21 7:03 code
  7:18 7:20 review
  zojirushi $20 

  This module defines an abstract theory of
  Incidence geoemetry.
*)

Require Import base.
Require Import List.
Require Import Bool.
Require Import Classical.
Require Import Description.

Module Incidence.

(* Represents an abstract Incidence geometry. *)
Structure Incidence : Type := incidence {
  (* Represents the set of points. *)
  point : Set;

  (* Asserts that equality between points is decideable. *)
  point_eq_dec : forall p q : point, { p = q } + { p <> q };

  (*
    Accepts three points and asserts that
    they are distinct.
  *)
  distinct
    :  point -> point -> point -> Prop
    := fun p q r => p <> q /\ p <> r /\ q <> r;

  (* Represents the set of lines. *)
  line : Set;

  (* Asserts that equality between lines is decideable. *)
  line_eq_dec : forall l m : line, { l = m } + { l <> m };

  (*
    Accepts a line and a point and returns true
    iff the point lies on the line.
  *)
  on : line -> point -> bool;

  (*
    Asserts that three points A, B  and c are
    collinear if there exists one line l such
    that all three points A, B, and C all lie
    on l.
    Accepts three points and returns true iff
  *)
  collinear
    :  point -> point -> point -> Prop
    := fun p q r : point
         => exists l : line, on l p = true /\ on l q = true /\ on l r = true; 

  (*
    Accepts three points and returns true iff
    they are not collinear.
  *)
  noncollinear
    :  point -> point -> point -> Prop
    := fun p q r => ~ collinear p q r;

  (*
    Two lines l and m are said to be parallel
    if there is no point P such that P lies on
    both  l and m.
  *)
  parallel
    :  line -> line -> Prop
    := fun l m => ~ exists p : point, on l p = true /\ on m p = true;

  (*
    Accepts two lines and returns true iff they
    are not parallel.
  *)
  nonparallel 
    :  line -> line -> Prop
    := fun l m => ~ parallel l m;

  (*
    Asserts that for every pair of distinct
    points P and Q there exists exactly one line
    l such that both P and Q lie on l.
  *)
  points_line_axiom : forall p q : point, p <> q -> exists! l : line, on l p = true /\ on l q = true;

  (*
    Asserts that for every line l there exists
    at least two distinct points P and Q such
    that both P and Q lie on l.
  *)
  line_points_axiom : forall l : line, exists p : point, exists q : point, p <> q -> on l p = true /\ on l q = true;

  (*
    Asserts that there exists three points that
    do not lie on any one line.

    Note: we must add the distinct condition or
    we could satisfy these axioms with a single
    point geometry.
  *)
  collinearity_axiom : exists p : point, exists q : point, exists r : point, distinct p q r /\ noncollinear p q r
}.

Arguments point_eq_dec {i} p q.
Arguments distinct {i} p q r.
Arguments line_eq_dec {i} l m.
Arguments on {i} l p.
Arguments collinear {i} p q r.
Arguments noncollinear {i} p q r.
Arguments parallel {i} l m.
Arguments nonparallel {i} l m.
Arguments points_line_axiom {i} p q H.
Arguments line_points_axiom {i} l.

Section Theorems.

(* Represents an arbitrary incidence geometry. *)
Variable g : Incidence.

(* Abbreviations. *)
Let point := point g.
Let line := line g.

(* Proves that at least two distinct lines exist. *)
Definition lines_ex
  :  exists l : line, exists m : line, l <> m
  := ex_ind (fun p : point =>
     ex_ind (fun q : point => 
     ex_ind (
       fun (r : point) (H : distinct p q r /\ noncollinear p q r)
         => ex_ind
              (fun (l : line) (H0 : unique (fun l => on l p = true /\ on l q = true) l)
                => ex_ind
                     (fun (m : line) (H1 : unique (fun m => on m q = true /\ on m r = true) m)
                       => ex_intro
                            (fun x => exists y : line, x = y -> False)
                            l
                            (ex_intro
                              (fun y => l = y -> False)
                              m
                              (fun H2 : l = m
                                => proj2 H
                                     (ex_intro
                                       (fun n => on n p = true /\ on n q = true /\ on n r = true)
                                       l
                                       (conj
                                         (proj1 (proj1 H0))
                                         (conj
                                           (proj2 (proj1 H0))
                                           (proj2 (proj1 H1)
                                            || on a r = true @a by H2)))))))
                     (points_line_axiom q r
                       (proj2 (proj2 (proj1 H)))))
               (points_line_axiom p q
                 (proj1 (proj1 H))))))
     (collinearity_axiom g).
 
Set Opaque lines_ex.

(*
  Proves that every pair of nonparallel lines
  have a point that lies on both lines.

  Note: this theorem uses the double negation
  rule from classical logic.
*)
Definition nonparallel_thm 
  :  forall l m : line, nonparallel l m -> exists p : point, on l p = true /\ on m p = true
  := fun l m (H : ~ ~ exists p, on l p = true /\ on m p = true)
       => NNPP (exists p, on l p = true /\ on m p = true) H.

Set Opaque nonparallel_thm.

(*
  If l and m are distinct nonparallel lines then
  there exists a unique point P such that P lies
  on both l and m.
*)
Definition unique_point_intersect_thm
  :  forall l m : line, l <> m -> nonparallel l m -> exists! p : point, on l p = true /\ on m p = true
  := fun l m H H0
       => let (p, H1) := nonparallel_thm l m H0 in
          ex_intro 
            (unique (fun p => on l p = true /\ on m p = true))
            p
            (conj H1
              (fun (q : point) (H2 : on l q = true /\ on m q = true)
                => sumbool_ind
                     (fun _ => p = q)
                     (fun H3 : p = q => H3)
                     (fun H3 : p <> q
                       => ex_ind
                            (fun n (H4 : unique (fun n => on n p = true /\ on n q = true) n)
                              => let H5
                                   :  on l p = true /\ on l q = true
                                   := conj (proj1 H1) (proj1 H2) in
                                 let H6
                                   :  on m p = true /\ on m q = true
                                   := conj (proj2 H1) (proj2 H2) in
                                 (False_ind
                                   (p = q)
                                   (H
                                     ((proj2 H4) m H6
                                      || a = m @a by <- (proj2 H4) l H5))))
                            (points_line_axiom p q H3))
                     (point_eq_dec p q))).

Set Opaque unique_point_intersect_thm.

(*
  Accepts two arguments, a point, p, and a list
  of points, qs, and asserts that none of the
  points in qs equal p.
*)
Definition points_different
  :  point -> list point -> Prop
  := fun p qs => Forall (fun q => q <> p) qs.

(*
  Accepts two arguments, a point, p, and a list
  of points, qs, and asserts that none of the
  points in qs equal p.
*)
Definition points_differentb
  :  point -> list point -> bool
  := fun p
       => list_rec
            (fun _ => bool)
            true
            (fun q qs (F : bool)
              => sumbool_rec
                   (fun _ => bool)
                   (fun H : q = p => false)
                   (fun H : q <> p => F)
                   (point_eq_dec q p)). 

(*
  Given a point, p, a list of points, qs, and
  a proof that p is not in qs, proves that p is
  not in the tail of qs either.
*)
Definition points_differentb_tail
  :  forall (p q : point) (qs : list point), points_differentb p (cons q qs) = true -> points_differentb p qs = true
  := fun p q qs
       => let t qs := points_differentb p qs = true in
          sumbool_ind
            (fun _ => t (cons q qs) -> t qs)
            (fun (H : points_differentb p qs = true) _
              => H)
            (fun (H : points_differentb p qs = false)
              => sumbool_ind
                   (fun H0 => sumbool_rec _ _ _ H0 = true -> t qs)
                   (fun (H0 : q = p) (H1 : false = true)
                     => False_ind
                          (t qs)
                          (diff_false_true H1))
                   (fun (H0 : q <> p) (H1 : t qs)
                     => (*
                          Note: technically we can just return H1, but that
                          would obscure the fact that this branch represents
                          a logical contradiction.
                        *)
                        False_ind
                          (t qs)
                          (diff_true_false
                            (H || a = false @a by <- H1)))
                   (point_eq_dec q p))
            (bool_dec0 (points_differentb p qs)).

Set Opaque points_differentb_tail.

(*
  Given a point, p, a list of points qs, and a
  proof that p is not in qs, proves that p does
  not equal the head of qs.
*)
Definition points_differentb_head
  :  forall (p q : point) (qs : list point), points_differentb p (cons q qs) = true -> q <> p
  := fun p q qs 
       => sumbool_ind
            (fun H 
              (*
                partial evaluation of points_differentb p (cons
                q qs) using: 

                Eval lazy beta delta [points_differentb
                list_rect list_rec] iota in (points_differentb
                p (cons q qs)).
              *)
              => (sumbool_rec _ _
                   (fun _ => (fix F (l : list point) : bool := _) qs)
                   H) = true -> q <> p)
            (fun (H : q = p) (H0 : false = true)
              => False_ind
                   (q <> p)
                   (diff_false_true H0))
            (fun (H : q <> p) _ => H)
            (point_eq_dec q p).

Set Opaque points_differentb_head.

(*
  Accepts a point, p, and a list of points,
  qs, where p is different from the head of
  qs and from every element in the tail of qs,
  and proves that p is not in qs.
*)
Definition points_differentb_ind
  :  forall (p q : point) (qs : list point), q <> p -> points_differentb p qs = true -> points_differentb p (cons q qs) = true
  := fun p q qs H H0
       => sumbool_ind
            (fun H1
              => sumbool_rec
                   (fun _ => bool)
                   (fun H : q = p => false)
                   (fun H : q <> p
                     => (fix F (rs : list point) : bool := _) qs)
                   H1 = true)
            (fun H1 : q = p
              => False_ind _ (H H1))
            (fun _
              => H0)
            (point_eq_dec q p).

Set Opaque points_differentb_ind.

(*
  Proves that points_differentb and
  points_different are equivalent.
*)
Definition points_differentb_thm
  :  forall (p : point) (qs : list point), points_differentb p qs = true <-> points_different p qs
  := fun p
       => list_ind
            (fun qs => points_differentb p qs = true <-> points_different p qs)
            (conj
              (fun _ => Forall_nil (fun q => q <> p))
              (fun _ => eq_refl true))
            (fun q qs (F : points_differentb p qs = true <-> points_different p qs)
              => conj
                   (fun H : points_differentb p (cons q qs) = true
                     => Forall_cons q
                          (points_differentb_head p q qs H)
                          ((proj1 F) (points_differentb_tail p q qs H) : Forall (fun q => q <> p) qs))
                   (fun H : points_different p (cons q qs)
                     => Forall_ind
                          (fun rs => points_differentb p rs = true)
                          (eq_refl true)
                          (fun r rs (H : r <> p) (H0 : points_different p rs)  (H1 : points_differentb p rs = true)  
                            => points_differentb_ind p r rs H
                                 (H1))
                          H)).

Set Opaque points_differentb_thm.

(*
  Given a point, p, a list of points qs, and a
  proof that p is not in qs, proves that p does
  not equal the head of qs.
*)
Definition points_different_head
  :  forall (p q : point) (qs : list point), points_different p (cons q qs) -> q <> p
  := fun p q qs H
       => points_differentb_head p q qs
            (proj2 (points_differentb_thm p (q :: qs)) H).

Set Opaque points_different_head.

(*
  Given a point, p, a list of points, qs, and
  a proof that p is not in qs, proves that p is
  not in the tail of qs either.
*)
Definition points_different_tail
  :  forall (p q : point) (qs : list point), points_different p (q :: qs) -> points_different p qs
  := fun p q qs H
       => proj1 (points_differentb_thm p qs)
            (points_differentb_tail p q qs
              (proj2 (points_differentb_thm p (q :: qs)) H)).

Set Opaque points_different_tail.

(*
  Accepts a point, p, and a list of points,
  qs, where p is different from the head of
  qs and from every element in the tail of qs,
  and proves that p is not in qs.
*)
Definition points_different_ind
  :  forall (p q : point) (qs : list point), q <> p -> points_different p qs -> points_different p (cons q qs)
  := fun p q qs H H0
       => (proj1 (points_differentb_thm p (q :: qs)))
            (points_differentb_ind p q qs H
              ((proj2 (points_differentb_thm p qs)) H0)).

Set Opaque points_different_ind.

(*
  Accepts a list of points and asserts that the
  list does not contain any duplicates.
*)
Inductive points_distinct : list point ->  Prop
  := points_distinct_nil  : points_distinct nil
  |  points_distinct_cons
    : forall (p : point) (qs : list point),
      points_different p qs ->
      points_distinct qs ->
      points_distinct (cons p qs).

(*
  Accepts a list of points and returns true iff
  the list does not contain any duplicates.
*)
Definition points_distinctb
  :  list point -> bool
  := list_rec
       (fun _ => bool)
       true
       (fun p qs F
         => points_differentb p qs && F).

(*
  Proves that points_distinctb and
  points_distinct are equivalent.
*)
Definition points_distinct_thm
  :  forall ps : list point, points_distinctb ps = true <-> points_distinct ps
  := list_ind
       (fun ps => points_distinctb ps = true <-> points_distinct ps)
       (conj
         (fun _ => points_distinct_nil)
         (fun _ => eq_refl true))
       (fun p ps (F : points_distinctb ps = true <-> points_distinct ps)
         => conj
              (fun H : points_distinctb (cons p ps) = true
                => let H0
                     :  points_differentb p ps = true /\ points_distinctb ps = true
                     := andb_prop
                          (points_differentb p ps)
                          (points_distinctb ps)
                          H in
                   points_distinct_cons p ps
                     ((proj1 (points_differentb_thm p ps)) (proj1 H0))
                     ((proj1 F) (proj2 H0)))
              (fun H : points_distinct (cons p ps)
                => points_distinct_ind
                     (fun rs => points_distinctb rs = true)
                     (eq_refl true)
                     (fun r rs (H0 : points_different r rs) (H1 : points_distinct rs) (H2 : points_distinctb rs = true)
                       => let H3
                            :  points_differentb r rs = true
                            := (proj2 (points_differentb_thm r rs)) H0 in
                          andb_true_intro (conj H3 H2))
                     (cons p ps)
                     H)).

Set Opaque points_distinct_thm.
 
(*
  Accepts two points and returns true iff they
  are equal.
*)
Definition point_eqb
  :  point -> point -> bool
  := fun p q
       => if point_eq_dec p q
            then true
            else false.

(*
  Accepts two points and returns true iff they
  are not equal.
*)
Definition point_neqb
  :  point -> point -> bool
  := fun p q
       => negb (point_eqb p q).

(*
  Proves that point_neqb is sound - i.e. that
  it returns true iff two points are different.
*)
Definition point_neqb_thm
  :  forall p q : point, point_neqb p q = true -> p <> q
  := fun p q
       => sumbool_ind
            (fun H => negb (if H then true else false) = true -> p <> q)
            (fun _ (H : false = true)
              => False_ind
                   (p <> q)
                   (diff_false_true H))
            (fun (H : p <> q) (_ : true = true)
              => H)
            (point_eq_dec p q).

Set Opaque point_negb_thm.
 
(* 
  Accepts a point, p, and a list of points,
  qs, and proves that, if every point in qs is
  different from p then `points_different p qs`.
*)
Definition points_different_In
  :  forall (p : point) (qs : list point), (forall q : point, In q qs -> q <> p) -> points_different p qs
  := fun p qs H
       => proj2 (Forall_forall (fun q => q <> p) qs) H.

Set Opaque points_different_In.

(*
  Accepts a point, p, and a list of points, qs,
  and proves that, if every point, q, in qs is
  not equal to p according to points_neqb then
  `points_different p qs`.
*)
Definition points_different_In_neqb
  :  forall (p : point) (qs : list point), (forall q : point, In q qs -> point_neqb q p = true) -> points_different p qs
  := fun p qs H
       => points_different_In p qs
            (fun q H0 => point_neqb_thm q p (H q H0)).

Set Opaque points_different_In_neqb.

(*
  Accepts a point, p, and a list, ps, and removes
  all instances of p from ps.
*)
Definition points_filter
  :  forall (p : point) (ps : list point), {qs : list point | points_different p qs}
  := fun p ps
       => let qs := filter (fun q => point_neqb q p) ps in
          let H
            :  points_different p qs
            := points_different_In_neqb p qs
                 (fun q H
                   => proj2 (proj1 (filter_In (fun q => point_neqb q p) q ps) H)) in
          exist
            (points_different p)
            qs
            H.

(*
  Accepts a point, p, and a list of distinct
  points, ps, and returns a list of distinct
  points that are different than p.
*)
Definition points_filterb
  :  forall (p : point) (ps : list point), {qs : list point | points_differentb p qs = true}
  := fun p ps
       => sig_rec
            (fun _ => {qs : list point | points_differentb p qs = true})
            (fun qs H
              => exist
                   (fun rs => points_differentb p rs = true)
                   qs
                   (proj2 (points_differentb_thm p qs) H))
            (points_filter p ps).

(*
  Accepts a list of distinct points and proves
  that the head of the list is different than
  all the remaining points and all the remaining
  points are distinct.
*)
Definition points_distinctb_inv
  :  forall (p : point) (ps : list point), points_distinctb (p :: ps) = true -> points_differentb p ps = true /\ points_distinctb ps = true
  := fun p ps (H : points_differentb p ps && points_distinctb ps = true)
       => andb_prop
            (points_differentb p ps) 
            (points_distinctb ps)
            H.

(*
  Accepts a list of distinct points and proves
  that the head of the list is different than
  all the remaining points and all the remaining
  points are distinct.
*)
Definition points_distinct_inv
  :  forall (p : point) (ps : list point), points_distinct (p :: ps) -> points_different p ps /\ points_distinct ps
  := fun p ps H
       => let H0
            :  points_distinctb (p :: ps) = true
            := proj2 (points_distinct_thm (p :: ps)) H in
          let H1
            :  points_differentb p ps = true /\ points_distinctb ps = true
            := points_distinctb_inv p ps H0 in
          conj
            (proj1 (points_differentb_thm p ps) (proj1 H1))
            (proj1 (points_distinct_thm ps) (proj2 H1)). 

(*
  Accepts a list of points and proves that the
  head of the list is different than all of the
  remaining points.
*)
Definition points_distinct_different
  :  forall (p : point) (ps : list point), points_distinct (p :: ps) -> points_different p ps
  := fun p ps H
       => proj1 (points_distinct_inv p ps H).

(*
  Accepts a list of distinct points and proves
  that the tail of the list is a list of distinct
  points.
*)
Definition points_distinct_tail
  :  forall (p : point) (ps : list point), points_distinct (p :: ps) -> points_distinct ps
  := fun p ps H
       => proj2 (points_distinct_inv p ps H).

(*
  Accepts two lists, xs and ys, and asserts that
  xs is a sublist of ys.
*)
Definition sublist
  :  forall A : Set, list A -> list A -> Prop
  := fun _ xs ys
       => forall x, In x xs -> In x ys.

Arguments sublist {A} xs ys.

(*
  Accepts two lists, xs and ys, where xs is a
  sublist of ys, and proves that every value
  that is not in ys is also not in xs.
*)
Definition sublist_thm
  :  forall (A : Set) (xs ys : list A), sublist xs ys -> forall z : A, ~ In z ys -> ~ In z xs
  := fun _ xs ys H z H0 (H1 : In z xs)
       => H0 (H z H1).

Arguments sublist_thm {A} xs ys H z H0 H1.

Set Opaque sublist_thm.

(*
  Accepts two lists of points, xs and ys, where
  xs is a sublist of ys, and proves that xs
  remains a sublist of ys even when we add the
  same element to both lists.
*)
Definition sublist_cons
  :  forall (xs ys : list point), sublist xs ys -> forall z : point, sublist (z :: xs) (z :: ys)
  := fun xs ys H z w (H0 : In w (z :: xs))
       => sumbool_ind
            (fun _ => In w (z :: ys))
            (fun H1 : z = w
              => or_introl (In w ys) H1)
            (fun H1 : z <> w
              => or_ind
                   (fun H2 : z = w
                     => False_ind (In w (z :: ys)) (H1 H2))
                   (fun H2 : In w xs
                     => or_intror (z = w) (H w H2))
                   H0)
            (point_eq_dec z w).

Set Opaque sublist_cons.

(*
  Proves that the tail of a list is a sublist
  of the list.
*)
Definition sublist_tail
  :  forall (A : Set) (x : A) (xs : list A), sublist xs (x :: xs)
  := fun _ x xs y (H : In y xs)
       => or_intror (x = y) H.

Arguments sublist_tail {A} x xs x0 H.

Set Opaque sublist_tail.

(*
  Proves that if a point is different than every
  point in a list it is not in the list.
*)
Definition points_different_not_In
  :  forall (p : point) (qs : list point), points_different p qs -> ~ In p qs
  := fun p
       => list_rect
            (fun qs => points_different p qs -> ~ In p qs)
            (fun _ (H : In p nil) => H)
            (fun q qs
              (F : points_different p qs -> ~ In p qs)
              (H : points_different p (q :: qs))
              (H0 : In p (q :: qs))
              => or_ind
                   (fun H1 : q = p
                     => points_different_head p q qs H H1)
                   (fun H1 : In p qs
                     => F (points_different_tail p q qs H) H1)
                   H0).

Set Opaque points_different_not_In.

(*
  Proves that if a point is not in a list,
  the point does not equal the first element in
  the list.
*)
Definition not_In_head
  :  forall (p q : point) (qs : list point), ~ In p (q :: qs) -> q <> p
  := fun p q qs H
       => sumbool_ind
            (fun _ => q <> p)
            (fun H0 : q = p
              => False_ind (q <> p)
                   (H (or_introl (In p qs) H0)))
            (fun H0 : q <> p
              => H0)
            (point_eq_dec q p).

Set Opaque not_In_head.

(*
  Proves that if a point is not in a list,
  then the point is not in the tail of the
  list either.
*)
Definition not_In_tail
  :  forall (p q : point) (qs : list point), ~ In p (q :: qs) -> ~ In p qs
  := fun p q qs H (H0 : In p qs)
       => H (or_intror (q = p) H0).

Set Opaque not_In_tail.

(*
  Proves that if a point, p, is not in a list,
  qs, then p is different than every point in qs.
*)
Definition points_not_In_different
  :  forall (p : point) (qs : list point), ~ In p qs -> points_different p qs
  := fun p
       => list_rect
            (fun qs => ~ In p qs -> points_different p qs)
            (fun H => Forall_nil (fun q => q <> p))
            (fun q qs
              (F : ~ In p qs -> points_different p qs)
              (H : ~ In p (q :: qs))
              => Forall_cons q
                   (not_In_head p q qs H)
                   (F (not_In_tail p q qs H))).

Set Opaque points_not_In_different.

(*
  Proves that if a point, p, is different than
  every point in a list, qs, then p is different
  than every point in every sublist of qs.
*)
Definition points_different_sublist
  :  forall (p : point) (qs : list point), points_different p qs -> forall (rs : list point), sublist rs qs -> points_different p rs
  := fun p qs H rs H0
       => points_not_In_different p rs
            (sublist_thm rs qs H0 p
              (points_different_not_In p qs H)).

Set Opaque points_different_sublist.

(*
  Proves that if a point, p, does not equal the
  first element in a list, qs, nor does it equal
  any of the remaining elements, p is not in qs.
*)
Definition points_not_In
  :  forall (p q : point) (qs : list point), q <> p -> ~ In p qs -> ~ In p (q :: qs)
  := fun p q qs H H0
       => or_ind 
            (fun H1 : q = p => H H1)
            (fun H1 : In p qs => H0 H1).

Set Opaque points_not_In.

(*
  Proves that given a point, p, and a list of
  points, ps, we can determine whether or not
  p is in ps.
*)
Definition points_In_dec
  :  forall (p : point) (ps : list point), {In p ps} + {~ In p ps}
  := fun p
       => list_rec
            (fun ps => {In p ps} + {~ In p ps})
            (right (In p nil)
              (fun H : In p nil => H))
            (fun q qs (F : {In p qs} + {~ In p qs})
              => sumbool_rec
                   (fun _ => {In p (q :: qs)} + {~ In p (q :: qs)})
                   (fun H : q = p
                     => left (~ In p (q :: qs))
                          (or_introl (In p qs) H))
                   (fun H : q <> p
                     => sumbool_rec
                          (fun _ => {In p (q :: qs)} + {~ In p (q :: qs)})
                          (fun H0 : In p qs
                            => left (~ In p (q :: qs))
                                 (or_intror (q = p) H0))
                          (fun H0 : ~ In p qs
                            => right (In p (q :: qs))
                                 (points_not_In p q qs H H0))
                          F)
                   (point_eq_dec q p)).
                          
Set Opaque points_In_dec.

(*
  Proves that given a point, p, and a list of
  points, ps, that contains p we can decideably
  determine whether or not p is the first element
  in ps or is contained in the tail of ps.
*)
Definition points_In_destr
  :  forall (p q : point) (qs : list point), In p (q :: qs) -> {q = p} + {In p qs}
  := fun p q qs H
       => sumbool_rec
            (fun _ => {q = p} + {In p qs})
            (fun H0 : q = p
              => left (In p qs) H0)
            (fun H0 : q <> p
              => sumbool_rec (* over points_In_dec *)
                   (fun _ => {q = p} + {In p qs})
                   (fun H1 : In p qs
                     => right (q = p) H1)
                   (fun H1 : ~ In p qs
                     => False_rec
                          ({q = p} + {In p qs})
                          (points_not_In p q qs H0 H1 H))
                   (points_In_dec p qs))
            (point_eq_dec q p).

Set Opaque points_In_destr.

(*
  Accepts a list of distinct points, ps, and
  a point, p, that is in ps, and returns the
  sublist of ps in which p has been removed.
*)
Definition points_distinct_remove
  :  forall ps : list point, points_distinct ps -> forall p : point, In p ps -> {qs : list point | points_different p qs /\ points_distinct qs /\ S (length qs) = length ps /\ sublist qs ps}
  := let T ps p qs := points_different p qs /\ points_distinct qs /\ S (length qs) = length ps /\ sublist qs ps in
     let U ps p := {qs : list point | T ps p qs} in
     let V ps := points_distinct ps -> forall p : point, In p ps -> U ps p in
     list_rect V
       (fun _ p (H : False) 
         => False_rec
              (U nil p)
              H)
       (fun p ps
         (F : V ps)
         (H0 : points_distinct (p :: ps))
         (q : point)
         (H1 : In q (p :: ps))
         => sumbool_rec (* over point_eq_dec *)
              (fun _ => U (p :: ps) q)
              (fun H2 : p = q
                => exist
                     (T (p :: ps) q)
                     ps
                     (conj
                       (points_distinct_different p ps H0
                         || points_different a ps @a by <- H2)
                       (conj
                         (points_distinct_tail p ps H0)
                         (conj
                           (eq_refl (S (length ps)))
                           (sublist_tail p ps)))))
              (fun H2 : p <> q
                => sumbool_rec (* over In q (p :: ps) *)
                     (fun _ => U (p :: ps) q)
                     (fun H3 : p = q
                       => False_rec
                            (U (p :: ps) q)
                            (H2 H3))
                     (fun H3 : In q ps
                       => let (rs, H4)
                            := F (points_distinct_tail p ps H0)
                                 q H3 in
                          exist
                            (T (p :: ps) q)
                            (p :: rs)
                            (conj
                              (points_different_ind q p rs H2 (proj1 H4))
                              (conj
                                (points_distinct_cons p rs
                                  (points_different_sublist p ps
                                    (points_distinct_different p ps H0)
                                    rs (proj2 (proj2 (proj2 H4))))
                                  (proj1 (proj2 H4)))
                                (conj
                                  (eq_refl (length (p :: ps))
                                    || S a = length (p :: ps) @a by (proj1 (proj2 (proj2 H4))) : S (length (p :: rs)) = length (p :: ps))
                                  (sublist_cons rs ps (proj2 (proj2 (proj2 H4))) p)))))
                     (points_In_destr q p ps H1))
              (point_eq_dec p q)).

(*
  Accepts a nonempty list of points, returns the
  first element in the list and the remaining elements.
*)
Definition points_list_destr
  :  forall ps : list point, length ps > 0 -> {x | ps = (fst x) :: (snd x)}
  := list_rect
       (fun ps => length ps > 0 -> {x | ps = (fst x) :: (snd x)})
       (fun H : 0 > 0
         => False_rec
              {x | nil = (fst x) :: (snd x)}
              (Gt.gt_irrefl 0 H))
       (fun p qs _ _
         => exist
              (fun x => (p :: qs) = (fst x) :: (snd x))
              (p, qs)
              (eq_refl (p :: qs))).

(*
  Accepts a line and a list of points and returns
  true iff all the ppints lie on the line.
*)
Definition points_on
  :  line -> list point -> bool
  := fun l => forallb (on l).

(*
  Accepts a list of points and asserts that they
  are collinear.
*)
Definition points_collinear
  :  list point -> Prop
  := fun ps => exists l : line, Forall (fun p => on l p = true) ps.

(*
  Accepts a line, l, and a list of points, ps,
  and proves that the points in ps are collinear
  if they all lie on l.
*)
Definition points_on_collinear
  :  forall (l : line) (ps : list point), points_on l ps = true -> points_collinear ps
  := fun l ps H
       => ex_intro
            (fun l => Forall (fun p => on l p = true) ps)
            l
            (proj2 (Forall_forall (fun p => on l p = true) ps)
              (proj1 (forallb_forall (on l) ps) H)).

(*
  Accepts a list of points and returns true iff
  they are collinear.

  Note: this function checks whether or not all
  of the points in ps lie on the line passing
  through the first two points.
*)
Definition points_collinearb
  :  forall ps : list point, points_distinct ps -> bool
  := list_rec
       (fun ps => points_distinct ps -> bool)
       (fun _ => true)
       (fun p ps _
         => list_rec
              (fun ps => points_distinct (cons p ps) -> bool)
              (fun _ => true)
              (fun q qs _ (H : points_distinct (cons p (cons q qs)))
                => let H0
                     :  q <> p
                     := points_differentb_head p q qs
                          (proj1
                            (andb_prop
                              (points_differentb p (cons q qs))
                              (points_distinctb (cons q qs))
                              (proj2 (points_distinct_thm (cons p (cons q qs))) H))) in
                   let (l, _)
                     := constructive_definite_description
                          (fun l => on l q = true /\ on l p = true)
                          (points_line_axiom q p H0) in
                   points_on l qs)
              ps).

(*
*)
Definition point_set
  :  Set
  := {ps : list point | points_distinct ps}.

(*
*)
Definition point_set_list
  :  point_set -> list point
  := fun ps => proj1_sig ps.

(*
*)
Definition point_set_distinct
  :  forall ps : point_set, points_distinct (point_set_list ps)
  := fun ps => proj2_sig ps.

(*
*)
Definition point_set_in
  :  point -> point_set -> Prop
  := fun p ps
       => In p (point_set_list ps).

(*
  Proves that we can always decide whether or
  not a point belongs to a point set.
*)
Definition point_set_in_dec
  :  forall (p : point) (ps : point_set), {point_set_in p ps} + {~ point_set_in p ps}
  := fun p ps
       => sumbool_rec
            (fun _ => {point_set_in p ps} + {~ point_set_in p ps})
            (fun H : point_set_in p ps
              => left (~ point_set_in p ps) H)
            (fun H : ~ point_set_in p ps
              => right (point_set_in p ps) H)
            (points_In_dec p (point_set_list ps)).

(*
*)
Definition point_set_subset
  :  point_set -> point_set -> Prop
  := fun ps qs
       => forall p : point, point_set_in p ps -> point_set_in p qs.

(*
*)
Definition point_set_disjoint
  :  point_set -> point_set -> Prop
  := fun ps qs
       => forall p : point, point_set_in p ps -> ~ point_set_in p qs.

(* Proves that point_set_disjoint is symmetrical. *)
Definition point_set_disjoint_sym
  :  forall ps qs : point_set, point_set_disjoint ps qs -> point_set_disjoint qs ps
  := fun ps qs (H : point_set_disjoint ps qs) p (H0 : point_set_in p qs) (H1 : point_set_in p ps)
       => H p H1 H0.

(*
*)
Definition point_set_eq
  :  point_set -> point_set -> Prop
  := fun ps qs
       => point_set_subset ps qs /\ point_set_subset qs ps.

(*
*)
Definition point_set_nil
  :  point_set
  := exist
       (fun ps => points_distinct ps)
       nil
       points_distinct_nil.

(*
*)
Definition point_set_difference
  :  point_set -> point_set -> point_set
  := fun ps qs
       => list_rec
            (fun 

(*
  TODO Fforgot to factor in the fact that point sets are distinct.
*)
Definition point_set_difference_strong
  :  forall ps qs : point_set, {rs : point_set | forall p : point, point_set_in p ps /\ ~ point_set_in p qs -> point_set_in p rs}
  := fun ps qs
       => let t ps rs := forall p : point, point_set_in p ps /\ ~ point_set_in p qs -> point_set_in p rs in
          let u ps := {rs | t ps rs} in
          list_rec u
            (exist
              (t nil)
              nil
              (fun p (H : In p nil /\ ~ point_set_in p qs)
                => False_ind
                     (t nil nil)
                     (and_ind
                       (fun (H0 : In p nil) _ => H0)
                       H)))
            (fun p rs (F : u rs)
              => sumbool_rec (* over in p qs *)
                   (fun _ => u (p :: rs))
                   (fun H : point_set_in p qs
                     => let (ss, H0) := F in
                        (* H0 : {ss : point_set | forall p : point, point_set_in p rs /\ ~ point_set_in p qs -> point_set_in p ss} *)
                        exist
                          (t (p :: rs))
                          ss
                          (fun q (H1 : point_set_in q (p :: rs) /\ ~ (point_set_in q qs))
                            => or_ind
                                 (fun H2 : p = q
                                   => False_ind
                                        (point_set_in q ss)
                                        (proj2 H1
                                          (H || point_set_in a qs @a by H2)))
                                 (fun H2 : In p rs
                                   => H0 (conj H2 (proj2 H1)))
                                 (proj1 H1)))
                   (fun H : ~ point_set_in p qs
                     => let (ss, H0) := F in
                        let ts := p :: ss in
                        exist
                          (t (p :: rs))
                          ts
                          (fun q (H1 : point_set_in q (p :: rs) /\ ~ (point_set_in q qs))
                            => or_ind
                                 (fun H2 : p = q
                                   => or_inrol (In q ss) H2)
                                 (fun H2 : In p rs
                                   => or_intror
                                        (p = q)
                                        (H0 (conj H2 (proj2 H1))))
                                 (proj1 H1)))
            (point_set_list ps)

(*

(*
*)
Definition point_set_singleton
  :  forall p : point -> {ps : point_set | unique (fun q : point => point_set_is q ps) p}

(*
*)
Definition point_set_add
  :  forall (p : point) (ps : point_set), {qs : list point | point_set_eq (point_set_difference qs ps) (point_
  := fun p ps
       => 

(*
*)
Definition point_set_remove
  :  forall (p : point) (ps : point_set), point_set_in p ps -> {qs : point_set | point_set_subset qs ps /\ ~ point_set_in p qs}

(*
*)
Definition point_set_replace_collinear
  :  forall 

(*
 TODO Give me a list of the points guaranteed to exist by collinearity_axiom.

 TODO  A function that returns two other points when given a point that are proved to be different and distinct 

  : {ps : list point
      | points_distinct ps /\
        length ps = 3

  points_others
    : forall p : point, exists q : point, exists r : point, distinct p q r /\ noncollinear p q r

  my goal was to call the collinearity axiom to get three distinct noncollinear points q r s
  then form these in a list ps := q :: r :: s :: nil
  prove the elements were distinct using points_distinct_cons
  use cases over points_In_dec
    if p is in ps
      call points_distinct_remove to get qs where qs is distinct and doesnt contain
      in this case we can prove that p :: qs is noncollinear but we'd need to deepen our theory (points_collinear + thms).
    if p is not in ps
      we can take the first two points and use points_different_head and points_different_tail iterative to prove they are all different 

  goal
    : forall (p : point) (ps : list point) (H : points_distinct ps), points_noncollinear ps -> forall H0 : In p ps, let (qs, H1) := points_distinct_remove ps H p H0 in points_distinct (p :: qs) /\ points_noncollinear (p :: ps)

*)
 
(*
*)
Definition points_others_noncollinear
  : forall p : point, exists q : point, exists r : point, distinct p q r /\ noncollinear p q r
  := fun p
       => ex_ind (fun q : point =>
          ex_ind (fun r : point =>
          ex_ind
            (fun (s : point) (H : distinct q r s /\ noncollinear q r s)
              => let ps := (q :: r :: s :: nil) in
                 sumbool_ind (* over In p ps *)
                   (fun _ => exists x : point, exists y : point, distinct p x y /\ noncollinear p x y)
                   (fun H0 : In p ps
                     => let (qs, H1)
                          := points_distinct_remove ps 
                               (* TODO: points_distinct ps *)
                               p H0 in
                        let rs := p :: qs in
                        let ((x, ss), H2)
                          := points_list_destr qs
                               (lt_0_Sn 1
                                || a > 0 @a by (eq_add_S (proj1 (proj2 (proj2 H1))) : length qs = 2)) in
                        let ((y, _), H3)
                          := points_list_destr ss
                               (lt_0_Sn 0
                                || a > 0 @a by 

(*
Check (eq_add_S (length (1 :: 2 :: 3 :: nil)) 3 (eq_refl (length (0 :: 1 :: 2 :: 3 :: nil))) : length (1 :: 2 :: 3 :: nil) = 3).
*)

(*
  Proves that at least one line passes through
  each point.
*)
Definition point_line_ex
  :  forall p : point, exists l : line, on l p = true
  := fun p
       => ex_ind (fun q =>
          ex_ind (fun r =>
          ex_ind (
            fun (s : point) (H : distinct q r s /\ noncollinear q r s)
              => let qs := q :: r :: s :: nil in
                 bool_dec0
                   (fun _ => exists l : line, on l p = true)
                   (fun H : points_differentb p qs = true
                     => let H0
                          :  p <> q
                          := points_differentb_head p q (r :: s :: nil) H in
                        ex_ind
                          (fun l (H1 : unique (fun l => on l p = true /\ on l q = true) l)
                            => ex_intro
                                 (fun l => on l p = true)
                                 l
                                 (proj1 (proj1 H1)))
                          (points_line_axiom p q H0))
                   (fun H : points_differentb p qs = false
                     => 

Definition points_collinearb_thm
  :  forall (ps : list point) (H : points_distinct ps), points_collinearb ps H = true <-> points_collinear ps
  := list_rec
       (fun ps => forall H : points_distinct ps, points_collinearb ps H = true <-> points_collinear ps)
       (conj
         (fun H _ => 
           (*
             points_collinearb nil H= true -> points_collinear nil
             _ -> exists l, Forall _ nil
           *)
           ex_ind
             (fun (l : line) (H0 : exists m : line, l <> m)
               => ex_intro
                    (fun l => Forall (fun p => on l p = true) nil)
                    l
                    (Forall_nil (fun p => on l p = true)))
             lines_ex)
         (fun _ _ => eq_refl true))
       (fun
         (p : point)
         (ps : list point)
         (H : forall H0 : points_distinct, points_collinearb ps H0 = true <-> points_collinear ps)
         (H0 : points_distinct (p::ps))
         => conj
              (*
                points_collinearb (p :: ps) = true -> points_collinear (p :: ps)
                _ => exists l : line, Forall (fun p => on l p = true) (p :: ps)
              *)
              (*
                two cases:
                  ps is nil
                    we use fact that every point lines on a line to get l. Then use Forall_cons with Forall_nil taking care of the recursive case.
                  ps is not nil
                  ps is q::qs
                    let line be the line through p and q
                    use points_collinearb_tail to prove that q::qs is collinear
                    use ex_ind to get m the line that all q::qs is on
                    use Forall_cons 
              *)
              


              (Forall_ind
                (fun qs => points_collinearb qs = true -> points_collinear qs)
                (fun _
                  => ex_intro
                       (fun l => Forall (fun p => on l p = true) nil)
                       l
                       (Forall_nil (fun p => on l p = true)))
                (fun q qs (H0 : 
           



           (* prove that there exists a  line, then use ex instantiation to get the line and Forall_nil for this proof case. *)


(*
  := fun ps
       => conj
            (fun H : points_collinearb ps = true
              => let H0
                   :  forall p : point, In p ps -> on l p = true
                   := proj1 (forallb_forall (on l) ps) H in
                 proj2 (Forall_forall (on l) ps) H0)
            (fun H : points_collinear ps
              => let H0
                   :  forall p : point, In p ps -> on l p = true
                   := proj1 (Forall_forall (fun p => on l p = true) ps) H in
                 proj2 (forallb_forall (on l) ps) H0).
*)
End Theorems.

End Incidence.
